 (**      TrashCart2.Pas by Alexander Myasnikov                     **)
 (**      TrashCart PRNG  based on Cartman-II block cipher          **)
 (**      Freeware for any use,  non-patented, opensource cipher    **)
 (**      WEB:       www.darksoftware.narod.ru                      **)
 (**      Cartman block cipher project information                  **)
 (**      WEB:       www.alexanderwdark.narod.ru                    **)


{$MODE DELPHI}

unit trashcart2ddp2;

interface


type
  TTrashCartPRNG = class (TObject)

  public
    // Recommended init routine (key points to key = keysize bytes + IV  = 16 bytes)
    // Check keysize constant in Cartman cipher unit
    procedure tc_init (key: pointer);

    // Init by longint seed by key expand routine using also some timers and counters
    procedure tc_seed (seed: longint);

    // Init by longint seed by key expand routine using also some timers and counters
    procedure tc_qpcseed (seed: longint);


    // Init by longint seed by key expand routine
    procedure tc_lgseed (seed: longint);


    // Generate len bytes and move to data buffer
    procedure tc_read (Data: pointer; len: integer);

    function tc_word: word;
    function tc_uint64: uint64;
    function tc_int64: int64;
    function tc_longword: longword;
    function tc_longint: longint;
    function tc_byte: byte;

    constructor Create;
  private
    keystream: array [0..15] of byte;
    ksleft:    integer;
    procedure tc_next ();
    function RDTSC: int64; register;
  end;

var
  TrashPRNG: TTrashCartPRNG = nil;

implementation


uses tccartman2ddp2, Windows;

type
  TLongIntKey = array [0.. ((16 + keysizebytes) div 4) - 1] of longint;

type
  PLongIntKey = ^TLongIntKey;

constructor TTrashCartPRNG.Create;
begin
  ksleft := 0;
end;

function TTrashCartPRNG.RDTSC: int64; register;
asm
  RDTSC
end;

procedure TTrashCartPRNG.tc_init (key: pointer);
begin
  tccartman_setkey(key);
  longword(key) := longword(key) + keysizebytes;
  move(key^, keystream, 16);
end;

procedure TTrashCartPRNG.tc_seed (seed: longint);
var
  key: pointer;
  i: integer;
begin
  GetMem(key, 16 + keysizebytes);

  PLongIntKey(key)[0] := seed;
  PLongIntKey(key)[1] := longint(GetTickCount());
  longword(key) := longword(key) + 8;
  PInt64(key)^  := RDTSC;
  longword(key) := longword(key) + 8;
  PLongIntKey(key)[0] := seed * 69069 + 1;
  for i := 1 to ((16 + keysizebytes) div 4) - 5 do
    PLongIntKey(key)[i] := PLongIntKey(key)[i - 1] * 69069 + 1;
  longword(key) := longword(key) - 16;
  tc_init(key);
  FreeMem(key);
end;

procedure TTrashCartPRNG.tc_qpcseed (seed: longint);
var
  key: pointer;
  i: integer;
begin
  GetMem(key, 16 + keysizebytes);
  PLongIntKey(key)[0] := seed;
  longword(key) := longword(key) + 4;
  QueryPerformanceCounter(PInt64(key)^);
  longword(key) := longword(key) + 8;
  PInt64(key)^  := RDTSC;
  longword(key) := longword(key) + 8;
  PLongIntKey(key)[0] := seed * 69069 + 1;
  for i := 1 to ((16 + keysizebytes) div 4) - 6 do
    PLongIntKey(key)[i] := PLongIntKey(key)[i - 1] * 69069 + 1;
  longword(key) := longword(key) - 20;
  tc_init(key);
  FreeMem(key);
end;


procedure TTrashCartPRNG.tc_lgseed (seed: longint);
var
  key: pointer;
  i: integer;
begin
  GetMem(key, 16 + keysizebytes);
  PLongIntKey(key)[0] := seed;

  for i := 1 to ((16 + keysizebytes) div 4) - 1 do
    PLongIntKey(key)[i] := PLongIntKey(key)[i - 1] * 69069 + 1;
  tc_init(key);
  FreeMem(key);
end;


procedure TTrashCartPRNG.tc_next ();
begin
  tccartman_crypt(@keystream);
end;


procedure TTrashCartPRNG.tc_read (Data: pointer; len: integer);
var
  buflen, i: integer;
begin

  if ksleft > 0 then
    begin

    for I := (16 - ksleft) to 15 do
      begin
      byte(Data^) := keystream[i];
      longword(Data) := longword(Data) + 1;
      Dec(ksleft);
      Dec(len);
      if len = 0 then
        break;

      end;

    end;


  if len > 0 then
    begin

    repeat

      if len >= 16 then
        buflen := 16
      else
        buflen := len;

      tc_next();

      for I := 0 to buflen - 1 do
        begin
        byte(Data^) := keystream[i];
        longword(Data) := longword(Data) + 1;
        end;


      Dec(len, buflen);

    until len <= 0;

    ksleft := 16 - buflen;
    end;

end;


function TTrashCartPRNG.tc_byte: byte;
begin
  tc_read(@Result, 1);
end;

function TTrashCartPRNG.tc_longint: longint;
begin
  tc_read(@Result, 4);
end;

function TTrashCartPRNG.tc_longword: longword;
begin
  tc_read(@Result, 4);
end;

function TTrashCartPRNG.tc_int64: int64;
begin
  tc_read(@Result, 8);
end;

function TTrashCartPRNG.tc_uint64: uint64;
begin
  tc_read(@Result, 8);
end;

function TTrashCartPRNG.tc_word: word;
begin
  tc_read(@Result, 2);
end;


initialization
  TrashPrng := TTrashCartPRNG.Create;

finalization
  TrashPrng.Free;

end.
